import copy
import os
import pickle
from GangaCore.Core.exceptions import BackendError
from GangaLHCb.Lib.LHCbDataset import LHCbDataset
from GangaGaudi.Lib.RTHandlers.GaudiDiracRunTimeHandler import GaudiDiracRunTimeHandler
from GangaGaudi.Lib.RTHandlers.RunTimeHandlerUtils import get_share_path, master_sandbox_prepare, sandbox_prepare, script_generator
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import dirac_inputdata, dirac_ouputdata, mangle_job_name, diracAPI_script_settings, API_nullifier
from GangaDirac.Lib.Backends.DiracUtils import result_ok
from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaDirac.Lib.Utilities.DiracUtilities import execute
from GangaCore.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns, getWNCodeForOutputPostprocessing
from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from GangaCore.GPIDev.Lib.File import FileBuffer, LocalFile
from GangaCore.GPIDev.Base.Proxy import addProxy, isType, stripProxy
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.util import unique
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaLHCb.Lib.RTHandlers.RTHUtils import getXMLSummaryScript, is_gaudi_child, lhcbdiracAPI_script_template, lhcbdirac_outputfile_jdl
logger = getLogger()


class LHCbGaudiDiracRunTimeHandler(GaudiDiracRunTimeHandler):

    def master_prepare(self, app, appmasterconfig):

        logger.debug("Master Prepare")
        inputsandbox, outputsandbox = master_sandbox_prepare(app, appmasterconfig, ['inputsandbox'])

        # add summary.xml
        outputsandbox += ['summary.xml', '__parsedxmlsummary__']

        logger.debug("Master Prepare LHCbGaudiDiracRunTimeHandler")

        return StandardJobConfig(inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):

        logger.debug("Prepare")

        inputsandbox, outputsandbox = sandbox_prepare(app, appsubconfig, appmasterconfig, jobmasterconfig)

        job = stripProxy(app).getJobObject()

        if job.inputdata:
            if not job.splitter:
                if len(job.inputdata) > 100:
                    raise BackendError("Dirac", "You're submitting a job to Dirac with no splitter and more than 100 files, please add a splitter and try again!")

        outputfiles = [this_file for this_file in job.outputfiles if isType(this_file, DiracFile)]

        data_str = 'import os\n'
        data_str += 'execfile(\'data.py\')\n'

        if hasattr(job, '_splitter_data'):
            data_str += job._splitter_data
        inputsandbox.append(FileBuffer('data-wrapper.py', data_str))

        input_data = []

        # Cant wait to get rid of this when people no-longer specify
        # inputdata in options file
        #######################################################################
        # splitters ensure that subjobs pick up inputdata from job over that in
        # optsfiles but need to take care of unsplit jobs
        if not job.master:
            share_path = os.path.join(get_share_path(app),
                                      'inputdata',
                                      'options_data.pkl')

            if not job.inputdata:
                if os.path.exists(share_path):
                    f = open(share_path, 'r+b')
                    job.inputdata = pickle.load(f)
                    f.close()

        #######################################################################

        # Cant wait to get rid of this when people no-longer specify
        # outputsandbox or outputdata in options file
        #######################################################################
        share_path = os.path.join(get_share_path(app),
                                  'output',
                                  'options_parser.pkl')

        if os.path.exists(share_path):
            #        if not os.path.exists(share_path):
            # raise GangaException('could not find the parser')
            f = open(share_path, 'r+b')
            parser = pickle.load(f)
            f.close()

            outbox, outdata = parser.get_output(job)

            from GangaCore.GPIDev.Lib.File import FileUtils
            from GangaCore.GPIDev.Base.Filters import allComponentFilters

            fileTransform = allComponentFilters['gangafiles']
            outdata_files = [fileTransform(this_file, None) for this_file in outdata if not FileUtils.doesFileExist(this_file, job.outputfiles)]
            job.non_copyable_outputfiles.extend([output_file for output_file in outdata_files if not isType(output_file, DiracFile)])
            outbox_files = [fileTransform(this_file, None) for this_file in outbox if not FileUtils.doesFileExist(this_file, job.outputfiles)]
            job.non_copyable_outputfiles.extend([outbox_file for outbox_file in outbox_files if not isType(outbox_file, DiracFile)])

            outputsandbox.extend([f.namePattern for f in job.non_copyable_outputfiles])

            outputsandbox.extend([f.namePattern for f in job.outputfiles if not isType(f, DiracFile)])
            outputsandbox = unique(outputsandbox)  # + outbox[:])
        #######################################################################

        input_data_dirac, parametricinput_data = dirac_inputdata(job.application)

        if input_data_dirac is not None:
            for f in input_data_dirac:
                if isType(f, DiracFile):
                    input_data.append(f.lfn)
                elif isType(f, str):
                    input_data.append(f)
                else:
                    raise ApplicationConfigurationError("Don't know How to handle anythig other than DiracFiles or strings to LFNs!")

        commandline = "python ./gaudipython-wrapper.py"
        if is_gaudi_child(app):
            commandline = 'gaudirun.py '
            commandline += ' '.join([str(arg) for arg in app.args])
            commandline += ' options.pkl data-wrapper.py'
        logger.debug('Command line: %s: ', commandline)

        gaudi_script_path = os.path.join(job.getInputWorkspace().getPath(), "gaudi-script.py")

        script_generator(gaudi_script_template(),
                         #remove_unreplaced = False,
                         outputfile_path=gaudi_script_path,
                         PLATFORM=app.platform,
                         COMMAND=commandline,
                         XMLSUMMARYPARSING=getXMLSummaryScript()  # ,
                         #OUTPUTFILESINJECTEDCODE = getWNCodeForOutputPostprocessing(job, '    ')
                         )

        #logger.debug( "input_data %s" % str( input_data ) )

        # We want to propogate the ancestor depth to DIRAC when we have
        # inputdata set
        if job.inputdata is not None and isType(job.inputdata, LHCbDataset):

            # As the RT Handler we already know we have a Dirac backend
            if type(job.backend.settings) is not dict:
                raise ApplicationConfigurationError('backend.settings should be a dict')

            if 'AncestorDepth' in job.backend.settings:
                ancestor_depth = job.backend.settings['AncestorDepth']
            else:
                ancestor_depth = job.inputdata.depth
        else:
            ancestor_depth = 0

        lhcbdirac_script_template = lhcbdiracAPI_script_template()

        lhcb_dirac_outputfiles = lhcbdirac_outputfile_jdl(outputfiles)

        # not necessary to use lhcbdiracAPI_script_template any more as doing our own uploads to Dirac
        # remove after Ganga6 release
        # NOTE special case for replicas: replicate string must be empty for no
        # replication
        dirac_script = script_generator(lhcbdirac_script_template,
                                        DIRAC_IMPORT='from LHCbDIRAC.Interfaces.API.DiracLHCb import DiracLHCb',
                                        DIRAC_JOB_IMPORT='from LHCbDIRAC.Interfaces.API.LHCbJob import LHCbJob',
                                        DIRAC_OBJECT='DiracLHCb()',
                                        JOB_OBJECT='LHCbJob()',
                                        NAME=mangle_job_name(app),
                                        APP_NAME=stripProxy(app).appname,
                                        APP_VERSION=app.version,
                                        APP_SCRIPT=gaudi_script_path,
                                        APP_LOG_FILE='Ganga_%s_%s.log' % (stripProxy(app).appname, app.version),
                                        INPUTDATA=input_data,
                                        PARAMETRIC_INPUTDATA=parametricinput_data,
                                        OUTPUT_SANDBOX=API_nullifier(outputsandbox),
                                        OUTPUTFILESSCRIPT=lhcb_dirac_outputfiles,
                                        # job.fqid,#outputdata_path,
                                        OUTPUT_PATH="",
                                        SETTINGS=diracAPI_script_settings(job.application),
                                        DIRAC_OPTS=job.backend.diracOpts,
                                        PLATFORM=app.platform,
                                        REPLICATE='True' if getConfig('DIRAC')['ReplicateOutputData'] else '',
                                        ANCESTOR_DEPTH=ancestor_depth,
                                        ## This is to be modified in the final 'submit' function in the backend
                                        ## The backend also handles the inputfiles DiracFiles ass appropriate
                                        INPUT_SANDBOX='##INPUT_SANDBOX##'
                                        )
        logger.debug("prepare: LHCbGaudiDiracRunTimeHandler")

        return StandardJobConfig(dirac_script,
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#

def gaudi_script_template():
    '''Creates the script that will be executed by DIRAC job. '''

    import inspect
    script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                                   'GaudiTemplate.py.template')

    from GangaCore.GPIDev.Lib.File import FileUtils
    script_template = FileUtils.loadScript(script_location, '')

    return script_template

#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: LCGSandboxCache.py,v 1.8 2009-03-12 12:17:31 hclee Exp $
###############################################################################
#
# LCG backend
#
# ATLAS/ARDA
#
# Date:   January 2007
import os
import os.path
import re

from urlparse import urlparse

from GangaCore.Core.GangaThread.MTRunner import MTRunner, Data, Algorithm

from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem

from GangaCore.Utility.logging import getLogger
from GangaCore.Utility.GridShell import getShell

from GangaCore.Lib.LCG.GridSandboxCache import GridSandboxCache, GridFileIndex
from GangaCore.Lib.LCG.Utility import get_md5sum

lcg_sandbox_cache_schema_datadict = GridSandboxCache._schema.inherit_copy().datadict
lcg_file_index_schema_datadict = GridFileIndex._schema.inherit_copy().datadict


class LCGFileIndex(GridFileIndex):

    """
    Data object containing LCG file index information. 

    @author: Hurng-Chun Lee 
    @contact: hurngchunlee@gmail.com
    """

    lcg_file_index_schema_datadict.update({
        'lfc_host': SimpleItem(defvalue='', copyable=1, doc='the LFC hostname'),
        'local_fpath': SimpleItem(defvalue='', copyable=1, doc='the original file path on local machine')
    })

    _schema = Schema(Version(1, 0), lcg_file_index_schema_datadict)
    _category = 'GridFileIndex'
    _name = 'LCGFileIndex'

    def __init__(self):
        super(LCGFileIndex, self).__init__()


class LCGSandboxCache(GridSandboxCache):

    '''
    Helper class for upladong/downloading/deleting sandbox files using lcg-cr/lcg-cp/lcg-del commands. 

    @author: Hurng-Chun Lee 
    @contact: hurngchunlee@gmail.com
    '''

    lcg_sandbox_cache_schema_datadict.update({
        'se': SimpleItem(defvalue='', copyable=1, doc='the LCG SE hostname'),
        'se_type': SimpleItem(defvalue='srmv2', copyable=1, doc='the LCG SE type'),
        'se_rpath': SimpleItem(defvalue='generated', copyable=1, doc='the relative path to the VO directory on the SE'),
        'lfc_host': SimpleItem(defvalue='', copyable=1, doc='the LCG LFC hostname'),
        'srm_token': SimpleItem(defvalue='', copyable=1, doc='the SRM space token, meaningful only when se_type is set to srmv2')
    })

    _schema = Schema(Version(1, 0), lcg_sandbox_cache_schema_datadict)
    _category = 'GridSandboxCache'
    _name = 'LCGSandboxCache'

    logger = getLogger()

    def __init__(self):
        super(LCGSandboxCache, self).__init__()
        self.protocol = 'lcg'

    def __setattr__(self, attr, value):
        if attr == 'se_type' and value not in ['', 'srmv1', 'srmv2', 'se']:
            raise AttributeError('invalid se_type: %s' % value)
        super(LCGSandboxCache, self).__setattr__(attr, value)

    def impl_upload(self, cred_req, files=[], opts=''):
        """
        Uploads multiple files to a remote grid storage.
        """

        shell = getShell(cred_req)

        if self.lfc_host:
            shell.env['LFC_HOST'] = self.lfc_host

        self.logger.debug(
            'upload file with LFC_HOST: %s', shell.env['LFC_HOST'])

        # the algorithm of uploading one file
        class MyAlgorithm(Algorithm):

            def __init__(self, cacheObj):
                Algorithm.__init__(self)
                self.cacheObj = cacheObj
                self.dirname = self.cacheObj.__get_unique_fname__()

            def process(self, file):
                # decide number of parallel stream to be used
                fsize = os.path.getsize(urlparse(file)[2])
                fname = os.path.basename(urlparse(file)[2])
                fpath = os.path.abspath(urlparse(file)[2])

                md5sum = get_md5sum(fpath, ignoreGzipTimestamp=True)
                nbstream = int((fsize * 1.0) / (10.0 * 1024 * 1024 * 1024))

                if nbstream < 1:
                    nbstream = 1  # min stream
                if nbstream > 8:
                    nbstream = 8  # max stream

                cmd = 'lcg-cr -t 180 --vo %s -n %d' % (
                    self.cacheObj.vo, nbstream)
                if self.cacheObj.se is not None:
                    cmd = cmd + ' -d %s' % self.cacheObj.se
                if self.cacheObj.se_type == 'srmv2' and self.cacheObj.srm_token:
                    cmd = cmd + ' -D srmv2 -s %s' % self.cacheObj.srm_token

                # specify the physical location
                cmd = cmd + \
                    ' -P %s/ganga.%s/%s' % (self.cacheObj.se_rpath,
                                            self.dirname, fname)

                # specify the logical filename
                # NOTE: here we assume the root dir for VO is /grid/<voname>
                lfc_dir = '/grid/%s/ganga.%s' % (
                    self.cacheObj.vo, self.dirname)
                if not self.cacheObj.__lfc_mkdir__(shell, lfc_dir):
                    self.cacheObj.logger.warning(
                        'cannot create LFC directory: %s' % lfc_dir)
                    return None

                cmd = cmd + ' -l %s/%s %s' % (lfc_dir, fname, file)
                rc, output, m = self.cacheObj.__cmd_retry_loop__(
                    shell, cmd, self.cacheObj.max_try)

                if rc != 0:
                    return False
                else:
                    match = re.search('(guid:\S+)', output)
                    if match:
                        guid = match.group(1)

                        fidx = LCGFileIndex()
                        fidx.id = guid
                        fidx.name = fname
                        fidx.md5sum = md5sum
                        fidx.lfc_host = self.cacheObj.lfc_host
                        fidx.local_fpath = fpath

                        self.__appendResult__(file, fidx)
                        return True
                    else:
                        return False

        myAlg = MyAlgorithm(cacheObj=self)
        myData = Data(collection=files)

        runner = MTRunner(
            name='sandboxcache_lcgcr', algorithm=myAlg, data=myData)
        runner.start()
        runner.join(-1)

        return runner.getResults().values()

    def impl_download(self, cred_req, files=[], dest_dir=None, opts=''):
        """
        Downloads multiple files from remote grid storages to 
        a local directory.
        """
        if not dest_dir:
            dest_dir = os.getcwd()
        self.logger.debug('download file to: %s', dest_dir)

        # the algorithm of downloading one file to a local directory
        class MyAlgorithm(Algorithm):

            def __init__(self, cacheObj, cred_req):
                Algorithm.__init__(self)
                self.cacheObj = cacheObj
                self.shell = getShell(cred_req)

            def process(self, file):

                guid = file.id
                lfn = file.attributes['local_fpath']
                lfc_host = file.attributes['lfc_host']
                fname = os.path.basename(urlparse(lfn)[2])

                self.shell.env['LFC_HOST'] = lfc_host
                self.cacheObj.logger.debug(
                    'download file with LFC_HOST: %s', self.shell.env['LFC_HOST'])

                cmd = 'lcg-cp -t %d --vo %s ' % (
                    self.cacheObj.timeout, self.cacheObj.vo)
                if self.cacheObj.se_type:
                    cmd += '-T %s ' % self.cacheObj.se_type
                cmd += '%s file://%s/%s' % (guid, dest_dir, fname)

                self.cacheObj.logger.debug('download file: %s', cmd)

                rc, output, m = self.cacheObj.__cmd_retry_loop__(
                    self.shell, cmd, self.cacheObj.max_try)

                if rc != 0:
                    return False
                else:
                    self.__appendResult__(file.id, file)
                    return True

        myAlg = MyAlgorithm(cacheObj=self, cred_req=cred_req)
        myData = Data(collection=files)

        runner = MTRunner(
            name='sandboxcache_lcgcp', algorithm=myAlg, data=myData)
        runner.start()
        runner.join(-1)

        return runner.getResults().values()

    def impl_delete(self, cred_req, files=[], opts=''):
        """
        Deletes multiple files from remote grid storages. 
        """

        # the algorithm of downloading one file to a local directory
        class MyAlgorithm(Algorithm):

            def __init__(self, cacheObj, cred_req):
                Algorithm.__init__(self)
                self.cacheObj = cacheObj
                self.shell = getShell(cred_req)

            def process(self, file):

                guid = file.id

                lfc_host = file.attributes['lfc_host']

                self.shell.env['LFC_HOST'] = lfc_host

                self.cacheObj.logger.debug(
                    'delete file with LFC_HOST: %s' % self.shell.env['LFC_HOST'])

                cmd = 'lcg-del -a -t 60 --vo %s %s' % (self.cacheObj.vo, guid)

                rc, output, m = self.cacheObj.__cmd_retry_loop__(
                    self.shell, cmd, self.cacheObj.max_try)

                if rc != 0:
                    return False
                else:
                    self.__appendResult__(file.id, file)
                    return True

        myAlg = MyAlgorithm(cacheObj=self, cred_req=cred_req)
        myData = Data(collection=files)

        runner = MTRunner(
            name='sandboxcache_lcgdel', algorithm=myAlg, data=myData)
        runner.start()
        runner.join(-1)

        # update the local index file
        del_files = runner.getResults().values()
        all_files = self.get_cached_files()

        left_files = []
        for f in all_files:
            if f not in del_files:
                left_files.append(f)

        self.impl_bookkeepUploadedFiles(left_files, append=False)

        return del_files

    # For GUID protocol
    def __lfc_mkdir__(self, shell, path, mode='775'):
        '''Creates a directory in LFC'''

        cmd = 'lfc-mkdir -p -m %s %s' % (mode, path)

        (rc, output, m) = self.__cmd_retry_loop__(shell, cmd, 1)

        if rc != 0:
            return False
        else:
            return True



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################

from GangaCore.GPIDev.Adapters.IPostProcessor import PostProcessException
from GangaCore.GPIDev.Adapters.IChecker import IChecker
from GangaCore.GPIDev.Schema import FileItem
from GangaCore.Utility.logging import getLogger
import commands
import copy
import os
import re


logger = getLogger()


class CustomChecker(IChecker):

    """User tool for writing custom check with Python.
       Make a file, e.g customcheck.py,
       In that file, do something like:

       def check(j):
           if j has passed:
               return True
           else: 
               return False


       When the job is about to be completed, Ganga will call this function and fail the job if False is returned.

    """
    _category = 'postprocessor'
    _name = 'CustomChecker'
    _schema = IChecker._schema.inherit_copy()
    _schema.datadict['module'] = FileItem(defvalue=None, doc='Path to a python module to perform the check.')
    _exportmethods = ['check']

    def check(self, job):
        if (self.module is None) or not self.module:
            raise PostProcessException( "No module is specified and so the check will fail.")
        if (self.module.name is None) or not os.path.isfile(self.module.name):
            raise PostProcessException("The module '%s' does not exist and so CustomChecker will do nothing!" % (self.module.name))

        result = None

        try:
            ns = {'job': job}
            execfile(self.module.name, ns)
            exec('_result = check(job)', ns)
            result = ns.get('_result', result)
        except Exception as e:
            raise PostProcessException('There was a problem with executing the module: %s, CustomChecker will do nothing!' % e)
        if result is not True and result is not False:
            raise PostProcessException('The custom check module did not return True or False, CustomChecker will do nothing!')
        if result is not True:
            logger.info('The custom check module returned False for job(%s)', job.fqid)
            return self.failure
        return self.success




##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: JobTree.py,v 1.2.4.4 2009-07-24 13:39:39 ebke Exp $
##########################################################################
import copy
import os
from GangaCore.Core.exceptions import GangaException
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Base.Proxy import isType, GPIProxyObjectFactory
from GangaCore.GPIDev.Schema import Schema, SimpleItem, Version
from GangaCore.GPIDev.Lib.Job import Job
from GangaCore.GPIDev.Lib.Registry.JobRegistry import RegistryKeyError
from GangaCore.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy, _wrap
from GangaCore.GPIDev.Base.Proxy import stripProxy, GPIProxyObject

import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()


class TreeError(GangaException):

    """JobTree class can raise a TreeError exception. 
    Exception numbers:
    1, Directory does not exist;
    2, Not a directory;
    3, Attemp to delete the root directory;
    4, Not a job object.
    """

    def __init__(self, number=-1, msg=None):
        Exception.__init__(self, msg)
        self.number = number


class JobTree(GangaObject):

    """The jobtree object in Ganga gives you the possibility to organise jobs
    in a directory structure. Jobs are stored in the jobtree by their index
    and you can think of it as a softlink.
    This also means that jobs can be placed in several folders at the same time.
    If you remove a job from the registry the references in the jobtreee will
    also automatically disappear (because registry calls cleanlinks() method).
    The jobtree is persisted in between Ganga sessions.
    """
    _schema = Schema(Version(1, 2), {'name': SimpleItem(''),
                                     'folders': SimpleItem({os.sep: {}}, protected=1, copyable=1),
                                     })

    _category = 'jobtrees'
    _name = 'JobTree'
    _exportmethods = ['exists', 'isdir', 'add', 'rm', 'cd',
                      'mkdir', 'ls', 'pwd', 'listdirs', 'listjobs',
                      'getjobs', 'find', 'cleanlinks', 'printtree']

    default_registry = 'jobs'
    _cwd = {}

    def __init__(self):
        super(JobTree, self).__init__()
        self._setRegistry(None)
        self.cwd([os.sep])
        self.folders = {os.sep: {}}

    def cwd(self, val=None):
        """This workaround is necessary to prevent overwriting 
        the current directory every time another session changes something"""
        if val is None:
            return JobTree._cwd.get(id(self), [os.sep])
        JobTree._cwd[id(self)] = val

    def __get_path(self, path=None):
        if path is None:
            return self.cwd()[:]
        else:
            pp = []
            if not os.path.isabs(path):
                d = os.path.join(self.pwd(), path)
            else:
                d = path

            d = os.path.normpath(d)

            while True:
                d, fn = os.path.split(d)
                if fn:
                    pp.insert(0, fn)
                else:
                    pp.insert(0, d)
                    break
            return pp

    def __folder_cd(self, path):

        folders = self.__get_folders()

        if path == []:
            return folders

        returnable_folder = folders
        top_level = ''

        ##  loop through elements in the path to get to the requested path from the user
        for _dir in path:
            ## catch thie exception
            if _dir not in returnable_folder:
                clean_path = os.path.join(*path)
                raise TreeError(1, "Directory %s does not exist in folder %s, accessing: %s" % (_dir, top_level, clean_path))

            ## 'cd' into the folder of interest
            returnable_folder = returnable_folder[_dir]

            if not isType(returnable_folder, type({})):
                clean_path = os.path.join(*path)
                raise TreeError(2, "%s not a directory, accessing: %s" % (_dir, clean_path))

            top_level = _dir

        return returnable_folder

    def __make_dir(self, path):

        _path = self.__get_path(path)

        returnable_folder = self.__folder_cd(_path[:-1])

        local_dir = _path[-1]

        if local_dir not in returnable_folder:
            returnable_folder[local_dir] = {}

        if not isType(returnable_folder[local_dir], type({})):
            clean_path = os.path.join(*_path)
            raise TreeError(2, "%s not a directory, accessing: %s" % (local_dir, clean_path))

        return returnable_folder[local_dir]

        ##  Perform some anity checking before returning the local folder structure
    def __get_folders(self):
        if not hasattr(self, 'folders'):
            setattr(self, 'folders', {os.sep: {}})
        f = self.folders
        if os.sep not in f:
            f[os.sep] = {}

        return f

    def __select_dir(self, path):

        ## sanitise the path and get an ordered list of the path directories
        _path = self.__get_path(path)

        returnable_folder = self.__folder_cd(_path)

        return returnable_folder

    def __remove_dir(self, path=None, dir=None):
        if dir is None:
            return
        if path is None:
            path = self.__get_path()
        else:
            if dir not in self.__get_folders()[path]:
                return
            else:
                del self.__get_folders()[path][dir]
        return


    ## Explicitly DO NOT copy self as we want one object per jobs repo through the GPI! - rcurrie

    def clone(self, ignora_atts=[]):
        return self

    def __deepcopy__(self, memo):
        return self

    def _copy(self):
        return self

    def _display(self, interactive=0):
        from GangaCore.Utility.ColourText import ANSIMarkup, NoMarkup
        if interactive:
            markup = ANSIMarkup()
        else:
            markup = NoMarkup()

        dirs = self.listdirs()
        cnt = len(dirs)
        ds = "\n[Folders]: %d\n" % cnt
        if cnt > 0:
            ds += "--------------\n"
        for d in dirs:
            ds += "%s\n" % d

        jobs = self.getjobs()
        ds += "[Jobs]: %d\n" % len(jobs)
        if len(jobs) > 0:
            ds += "--------------\n"
            ds += jobs._display(interactive)

        return ds

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text('jobtree...')
            return
        p.text(self._display(interactive=True))

    def _proxy_display(self, interactive=1):
        return self._display(interactive=interactive)

    def exists(self, path):
        """Checks wether the path exists or not.
        """
        _path = self.__get_path(path)
        try:
            folder = self.__folder_cd(_path[:-1])
            local_dir = path[-1]
            if local_dir in path:
                return True
            else:
                return False
        except Exception, err:
            clean_path = os.path.join(*path)
            logger.debug('Error getting path: %s' % clean_path)
            logger.debug('%s' % err)
            return False
        return True

    def isdir(self, path):
        """Checks wether the path points to a folder or not.
        """
        try:
            self.__select_dir(path)
        except TreeError:
            return False
        return True

    def add(self, job, path=None):
        """Adds job to the job tree into the current folder.
        If path to a folder is provided as a parameter than adds job to that folder.
        """
        try:
            job = stripProxy(job)

            mydir = self.__select_dir(path)

            if isType(job, Job):
                mydir[job.getFQID('.')] = job.getFQID('.')  # job.id
            elif isType(job, JobRegistrySlice):
                for sliceKey in job.objects.iterkeys():
                    mydir[sliceKey] = sliceKey
            elif isType(job, list):
                for element in job:
                    mydir[element.id] = element.id
            else:
                raise TreeError(4, "Not a job/slice/list object")

            self._setDirty()
        except Exception, err:
            logger.error("Error: %s" % err)
            raise
        finally:
            self._releaseSessionLockAndFlush()

    def rm(self, path):
        """Removes folder or job in the path.
        To clean all use /* as a path.
        """
        try:
            path = str(path)
            pp = self.__get_path(path)
            if len(pp) > 1:
                dpath = os.path.join(*pp[:-1])
                f = self.__select_dir(dpath)
                if pp[-1] != '*':
                    try:
                        self.__remove_dir(path=dpath, dir=pp[-1])
                    except KeyError:
                        raise TreeError(1, "%s does not exist" % pp[-1])
                else:
                    for k in self.__get_folders().keys():
                        del self.__get_folders()[k]
            else:
                raise TreeError(3, "Can not delete the root directory")
            self._setDirty()
        finally:
            self._releaseSessionLockAndFlush()

    def mkdir(self, path):
        """Makes a folder. If any folders in the path are missing they will be created as well.
        """
        try:
            self.__make_dir(path)
            self._setDirty()
        finally:
            self._releaseSessionLockAndFlush()

    def cd(self, path=os.sep):
        """Changes current directory.
        If path is not provided, than switches to the root folder.
        """
        try:
            self.__select_dir(path)
            self.cwd(self.__get_path(path))
            self._setDirty()
        finally:
            self._releaseSessionLockAndFlush()

    def ls(self, path=None):
        """Lists content of current folder or folder in the path if the latter is provided.
        The return value is a dictionary of the format
        {'folders':[<list of folders>], 'jobs':[<list of job ids>]}.
        """
        _path = self.__get_path(path)
        top_level = self.__folder_cd(_path)
        res = {'folders': [], 'jobs': []}
        for i in top_level:
            if isType(top_level[i], dict):
                res['folders'].append(i)
            else:
                res['jobs'].append(i)
        return res

    def listdirs(self, path=None):
        """Lists all subfolders in current folder or folder in the path if the latter is provided.
        """
        return self.ls(path)['folders']

    def listjobs(self, path=None):
        """Lists ids of all jobs in current folder or folder in the path if the latter is provided.
        """
        return self.ls(path)['jobs']

    def pwd(self):
        """Returns current folder"""
        return os.path.join(*self.cwd())

    def getjobs(self, path=None):
        """Gives list of all jobs (objects) referenced in current folder
        or folder in the path if the latter is provided.
        """
        # jobslice
        ##res = []
        res = JobRegistrySlice("")
        registry = self._getRegistry()
        do_clean = False
        if registry is not None:
            try:
                registry = registry._parent
            except:
                GangaCore.Utility.logging.log_unknown_exception()
                pass
            path = os.path.join(*self.__get_path(path))
            res.name = "jobs found in %s" % path
            cont = self.ls(path)
            for i in cont['jobs']:
                try:
                    try:
                        id = int(i)
                        j = registry[id]
                    except ValueError:
                        j = registry[int(i.split('.')[0])].subjobs[int(i.split('.')[1])]
                except RegistryKeyError, ObjectNotInRegistryError:
                    do_clean = True
                else:
                    res.objects[j.id] = j
        if do_clean:
            self.cleanlinks()
        return _wrap(res)

    def find(self, id, path=None):
        """For a job with given id tries to find all references in the job tree.
        The return value is a list of found paths.
        """

        if isType(id, Job):
            id = stripProxy(id).getFQID('.')
        if isType(id, GPIProxyObject):
            id = stripProxy(id)

        pp = self.__get_path(path)
        tp = os.path.join(*pp)

        top_level = self.__folder_cd(pp)

        search_dirs = []
        found_items = []
        for _item in top_level.keys():
            if isinstance(top_level[_item], dict):
                search_dirs.append(_item)
            else:
                if top_level[_item] == id:
                    found_items.append(tp)

        result = []

        for item in found_items:
            result.append(item)

        for dir in search_dirs:
            new_path = os.path.join(tp, dir)
            for found in self.find(id, new_path):
                result.append(found)

        return result

    def cleanlinks(self, path=os.sep):
        """Removes all references for the jobs not present in the registry.
        Normally you don't need to call this method since it is called automatically whenever
        job is deleted from the registry.
        """
        registry = self._getRegistry()
        if registry is not None:
            registry = registry._parent
            pp = self.__get_path(path)
            fc = self.__folder_cd(pp)

            #print("self._getRegistry(): %s" % stripProxy(self)._getRegistry())

            for i in fc:
                if isType(fc[i], type({})):
                    pass
                    self.cleanlinks(os.path.join(path, i))
                else:
                    try:
                        try:
                            _id = int(fc[i])
                            j = registry[_id]
                        except ValueError:
                            jid = fc[i].split('.')
                            j = registry[int(jid[0])].subjobs[int(jid[1])]
                    except RegistryKeyError, ObjectNotInRegistryError:
                        #try:
                        self.__remove_dir(path=path, dir=i)
                        self._setDirty()
                        #except ObjectNotInRegistryError as err:
                        #    logger.debug("Object: %s Not in Reg: %s" % (_id, err))
                        #    pass
                        try:
                            pass
                        finally:
                            try:
                                self._releaseSessionLockAndFlush()
                            except ObjectNotInRegistryError as err:
                                logger.debug("Object: %s Not in Reg: %s" % (_id, err))
                                pass

    def printtree(self, path=None):
        """Prints content of the job tree in a well formatted way.
        """
        def printdir(path, indent):
            cont = self.ls(path)
            ind = '    ' * indent
            ds = ''
            for i in cont['jobs']:
                ds += ind + '|-%s\n' % i
            for i in cont['folders']:
                ds += ind + '|-[%s]\n' % i
                ds += printdir(os.path.join(path, i), indent + 1)
            return ds
        path = os.path.join(*self.__get_path(path))
        bn = os.path.basename(path)
        if bn == '':
            bn = os.sep
        ds = '  [%s]\n' % bn
        logger.info(ds + printdir(path, 1))


class _proxy_display(object):

    def __get__(self, obj, cls):
        if obj is None:
            return stripProxy(cls)._proxy_display
        return stripProxy(obj)._proxy_display

class _copy(object):

    def __get__(self, obj, cls):
        if obj is None:
            return stripProxy(cls)._copy
        return stripProxy(obj)._copy


JobTree.__str__ = JobTree._display




from __future__ import absolute_import


def execInThread(target, args=None, kwargs=None, timeout=None,
                 waitFlag=False, lock=(None, None),
                 callBackFunc=None):

    if args is None:
        args = ()
    if kwargs is None:
        kwargs = {}

    def _child(target, args, kwargs):
        try:
            if lock[0]:
                if lock[1] is None:
                    lock[0].acquire()
                else:
                    lock[0].acquire(lock[1])
            result = target(*args, **kwargs)
        finally:
            if lock[0]:
                lock[0].release()
        if callBackFunc:
            callBackFunc(result)
        return

    if not callable(target):
        return False

    import threading
    childThread = threading.Thread(target=_child(target, args, kwargs))
    childThread.start()

# ------------------------

from .util import GenericWrapper


class SynchronisedObject(GenericWrapper):
    __slots__ = list()
    def __init__(self, obj, ignore=None, lock=None):
        if ignore is None:
            ignore = ()
        if lock is None:
            import threading
            lock = threading.RLock()
        GenericWrapper.__init__(self, obj, lock.acquire, lock.release, ignore)



from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.GPIDev.Lib.Tasks.IUnit import IUnit
from GangaCore.GPIDev.Lib.Job.Job import JobError
from GangaCore.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Base.Proxy import addProxy, stripProxy
from GangaLHCb.Lib.Splitters.SplitByFiles import SplitByFiles
from GangaCore.GPIDev.Lib.Tasks.common import makeRegisteredJob, getJobByID
from GangaCore.Utility.logging import getLogger

logger = getLogger()

class LHCbUnit(IUnit):
    _schema = Schema(Version(1, 0), dict(IUnit._schema.datadict.items() + {
        'input_datset_index': SimpleItem(defvalue=-1, protected=1, hidden=1, doc='Index of input dataset from parent Transform', typelist=["int"]),
    }.items()))

    _category = 'units'
    _name = 'LHCbUnit'
    _exportmethods = IUnit._exportmethods + []

    def createNewJob(self):
        """Create any jobs required for this unit"""
        import copy
        j = makeRegisteredJob()
        j.backend = self._getParent().backend.clone()
        j.application = self._getParent().application.clone()
        if self.inputdata:
            j.inputdata = self.inputdata.clone()

        j.inputfiles = copy.deepcopy(self._getParent().inputfiles)

        trf = self._getParent()
        task = trf._getParent()
        j.inputsandbox = self._getParent().inputsandbox

        j.outputfiles = copy.deepcopy(self._getParent().outputfiles)
        if len(self._getParent().postprocessors.process_objects) > 0:
            j.postprocessors = copy.deepcopy(
                addProxy(self._getParent()).postprocessors)

        if trf.splitter:
            j.splitter = trf.splitter.clone()

            # change the first event for GaussSplitter
            from GangaLHCb.Lib.Splitters.GaussSplitter import GaussSplitter
            if isType(trf.splitter, GaussSplitter):
                events_per_unit = j.splitter.eventsPerJob * \
                    j.splitter.numberOfJobs
                j.splitter.firstEventNumber += self.getID() * events_per_unit

        else:
            j.splitter = SplitByFiles()

        return j

    def checkMajorResubmit(self, job):
        """check if this job needs to be fully rebrokered or not"""
        return False

    def majorResubmit(self, job):
        """perform a major resubmit/rebroker"""
        super(LHCbUnit, self).majorResubmit(job)

    def reset(self):
        """Reset the unit completely"""
        super(LHCbUnit, self).reset()

    def updateStatus(self, status):
        """Update status hook"""

        # check for input data deletion of chain data
        if status == "completed" and self._getParent().delete_chain_input and len(self.req_units) > 0:

            # the inputdata field *must* be filled from the parent task
            # NOTE: When changing to inputfiles, will probably need to check
            # for any specified in trf.inputfiles

            # check that the parent replicas have been copied by checking
            # backend status == Done
            job_list = []
            for req_unit in self.req_units:
                trf = self._getParent()._getParent().transforms[
                    int(req_unit.split(":")[0])]
                req_unit_id = req_unit.split(":")[1]

                if req_unit_id != "ALL":
                    unit = trf.units[int(req_unit_id)]
                    job_list.append(getJobByID(unit.active_job_ids[0]))
                else:
                    for unit in trf.units:
                        job_list.append(getJobByID(unit.active_job_ids[0]))

            for j in job_list:
                if j.subjobs:
                    for sj in j.subjobs:
                        if sj.backend.status != "Done":
                            return
                else:
                    if j.backend.status != "Done":
                        return

            job = getJobByID(self.active_job_ids[0])
            for f in job.inputdata.files:
                # check for an lfn
                if hasattr(f, "lfn"):
                    fname = f.lfn
                else:
                    fname = f.namePattern

                logger.warning(
                    "Removing chain inputdata file '%s'..." % fname)
                f.remove()

        super(LHCbUnit, self).updateStatus(status)

    def checkForSubmission(self):
        """Additional checks for unit submission"""

        # call the base class
        if not super(LHCbUnit, self).checkForSubmission():
            return False

        return True



################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 23/04/2014
################################################################################
"""@package ND280Skimmer
Ganga module with classes to skim from reco files a set of events listed in a CSV file.
"""

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

from os.path import isfile

import os, shutil, commands, re, time
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class ND280RecoSkimmer(IPrepareApp, IPrepareApp):
    """
    ND280RecoSkimmer application to skim reco files from a list of run, subrun and event numbers in a CSV file.

    The required input for this module are:
        app.csvfile = '/home/me/myCSVfile.txt'
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
        app.outputfile = 'myFantasticResults.root'
    
    """
    _schema = Schema(Version(1,1), {
        'csvfile' : SimpleItem(defvalue=None,doc='CSV file containing "run,subrun,event" numbers.', typelist=['str','type(None)']),
        'cmtsetup' : SimpleItem(defvalue=None,doc='Setup script in bash to set up cmt and the cmt package of the executable.', typelist=['str','type(None)']),
        'outputfile' : SimpleItem(defvalue=None,doc='Output filename or filenames. Takes a string or a list of strings.', typelist=['str','type(None)','list']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'ND280RecoSkimmer'
    _scriptname = None
    _exportmethods = []
    _GUIPrefs = [ { 'attribute' : 'csvfile', 'widget' : 'String' },
                  { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                  { 'attribute' : 'outputfile', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'csvfile', 'widget' : 'String' },
                          { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                          { 'attribute' : 'outputfile', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    def __init__(self):
        super(ND280RecoSkimmer,self).__init__()


    def configure(self,masterappconfig):
        
        exefile = 'skimFromCSV.exe'
        exe = 'skimFromCSV.exe'
        # exe = '/'.join([os.getenv("RECONUTILSROOT"),os.getenv("CMTCONFIG"),exefile])
        # if not isfile(exe):
        #   raise ApplicationConfigurationError(None,'Cannot find executable '+exe)

        job = self.getJobObject()

        if self.cmtsetup is None:
          raise ApplicationConfigurationError('No cmt setup script given.')
        if not isfile(self.cmtsetup):
          raise ApplicationConfigurationError('Cannot find cmt setup script '+self.cmtsetup)

        # Copy CSV file to inputdir. Done in splitter for subjobs.
        if not isfile(self.csvfile):
          raise ApplicationConfigurationError('Cannot find CSV file '+self.csvfile)
        from shutil import copy
        tmpcsv = os.path.join(job.inputdir, os.path.basename(self.csvfile))
        if not os.path.exists(tmpcsv):
          copy(self.csvfile, job.inputdir)
        self.csvfile = tmpcsv

        args = []

        args.append('-O')
        args.append('file='+self.csvfile)

        if self.outputfile is None:
          raise ApplicationConfigurationError('No output file given. Fill the outputfile variable.')

        args.append('-o')
        args.append(self.outputfile)

        # Read the CSV file
        csvfile = open(self.csvfile, 'rb')
        run_subrun = []
        for line in csvfile:
            if line[0] == '#':
              continue
            row = line.split(",")
            if len(row) < 3:
              print "Ignoring badly-formatted line:", ",".join(row)
              continue

            r_sr = "%(run)08d-%(subrun)04d" % { "run" : int(row[0]), "subrun" : int(row[1]) }
            if r_sr not in run_subrun:
              run_subrun.append(r_sr)

        # So get the list of filenames get_dataset_filenames() and create a file containing the list of files and put it in the sandbox
        if job.inputdata is None:
          raise ApplicationConfigurationError('The inputdata variable is not defined.')
        rawFileList = job.inputdata.get_dataset_filenames()
        if len(rawFileList) < 1:
          raise ApplicationConfigurationError('No input data file given.')

        fileList = []
        for r_sr in run_subrun:
          for rfile in rawFileList:
            if rfile.find(r_sr) > -1:
              fileList.append(rfile)
              continue
        if not len(fileList):
          raise ApplicationConfigurationError('No file matching the run_subrun in the CSV file %s.' % self.csvfile)
        args.extend(fileList)

        argsStr = ' '.join(args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        script += 'source '+self.cmtsetup+'\n'
        script += '${RECONUTILSROOT}/${CMTCONFIG}/'+exe+' '+argsStr+'\n'
        # Little trick to be able to control the final destination
        # of the subjob's CSV file with SandboxFile or MassStorageFile
        if job.master is not None:
            script += 'cp %s .' % self.csvfile

        from GangaCore.GPIDev.Lib.File import FileBuffer

        if exefile.find('.exe') > -1:
            scriptname = exefile.replace('.exe', '.sh')
        else:
            scriptname = exefile + '.sh'
        job.getInputWorkspace().writefile(FileBuffer(scriptname,script),executable=1)

        self._scriptname = job.inputdir+scriptname

        return (None,None)



config = getConfig('defaults_ND280RecoSkimmer') #_Properties


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        return StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('ND280RecoSkimmer','LSF', RTHandler)
allHandlers.add('ND280RecoSkimmer','Local', RTHandler)
allHandlers.add('ND280RecoSkimmer','PBS', RTHandler)
allHandlers.add('ND280RecoSkimmer','SGE', RTHandler)
allHandlers.add('ND280RecoSkimmer','Condor', RTHandler)
allHandlers.add('ND280RecoSkimmer','LCG', LCGRTHandler)
allHandlers.add('ND280RecoSkimmer','gLite', gLiteRTHandler)
allHandlers.add('ND280RecoSkimmer','TestSubmitter', RTHandler)
allHandlers.add('ND280RecoSkimmer','Interactive', RTHandler)
allHandlers.add('ND280RecoSkimmer','Batch', RTHandler)
allHandlers.add('ND280RecoSkimmer','Cronus', RTHandler)
allHandlers.add('ND280RecoSkimmer','Remote', LCGRTHandler)
allHandlers.add('ND280RecoSkimmer','CREAM', LCGRTHandler)




################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 10/02/2014
################################################################################
"""@package ND280Executable
This module is designed to run any ND280 executable accessible in the $PATH environment variable.
"""

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

import os, shutil, commands, re, time
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

class ND280Executable(IPrepareApp, IPrepareApp):
    """
    ND280Executable application running any ND280 executables.

    The required input for this module are:
        app.exe = 'RunOARecon.exe'
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
        app.outputfile = 'myFantasticResults.root'

    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to analyze only 10 events:
        app.args = ['-n',10]
    """
    _schema = Schema(Version(1,1), {
        'exe' : SimpleItem(defvalue=None,typelist=['str', 'type(None)'],comparable=1,doc='A path (string) or a File object specifying an executable.'), 
        'args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'cmtsetup' : SimpleItem(defvalue=None,doc='Setup script in bash to set up cmt and the cmt package of the executable.', typelist=['str','type(None)']),
        'outputfile' : SimpleItem(defvalue=None,doc='Output filename or filenames. Takes a string or a list of strings.', typelist=['str','type(None)','list']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'ND280Executable'
    _scriptname = None
    _exportmethods = []
    _GUIPrefs = [ { 'attribute' : 'exe', 'widget' : 'File' },
                  { 'attribute' : 'args', 'widget' : 'String_List' },
                  { 'attribute' : 'outputfile', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'exe', 'widget' : 'File' },
                          { 'attribute' : 'args', 'widget' : 'String_List' },
                          { 'attribute' : 'outputfile', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    def __init__(self):
        super(ND280Executable,self).__init__()


    def configure(self,masterappconfig):
        
        self.args = convertIntToStringArgs(self.args)

        job = self.getJobObject()

        if self.cmtsetup is None:
          raise ApplicationConfigurationError('No cmt setup script given.')

        # Need to handle the possibility of multiple output files !
        # setup the output file
        for arg in self.args:
          if arg == '-o':
            raise ApplicationConfigurationError('Option "-o" given in args. You must use the outputfile variable instead, even if you have multiple output files.')

        if self.outputfile is None:
          raise ApplicationConfigurationError('No output file given. Fill the outputfile variable.')
        else:
          if type(self.outputfile) == type([]):
            for OutFi in self.outputfile:
              self.args.append('-o')
              self.args.append(OutFi)
          else:
            self.args.append('-o')
            self.args.append(self.outputfile)


        # So get the list of filenames get_dataset_filenames() and create a file containing the list of files and put it in the sandbox
        if job.inputdata is None:
          raise ApplicationConfigurationError('The inputdata variable is not defined.')
        fileList = job.inputdata.get_dataset_filenames()
        if len(fileList) < 1:
          raise ApplicationConfigurationError('No input data file given.')
        self.args.extend(fileList)

        argsStr = ' '.join(self.args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        script += 'source '+self.cmtsetup+'\n'
        script += self.exe+' '+argsStr+'\n'

        from GangaCore.GPIDev.Lib.File import FileBuffer

        if self.exe.find('.exe') > -1:
            scriptname = self.exe.replace('.exe', '.sh')
        else:
            scriptname = self.exe + '.sh'
        job.getInputWorkspace().writefile(FileBuffer(scriptname,script),executable=1)

        self._scriptname = job.inputdir+scriptname

        return (None,None)



config = getConfig('defaults_ND280Executable') #_Properties
config.options['exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        return StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('ND280Executable','LSF', RTHandler)
allHandlers.add('ND280Executable','Local', RTHandler)
allHandlers.add('ND280Executable','PBS', RTHandler)
allHandlers.add('ND280Executable','SGE', RTHandler)
allHandlers.add('ND280Executable','Condor', RTHandler)
allHandlers.add('ND280Executable','LCG', LCGRTHandler)
allHandlers.add('ND280Executable','gLite', gLiteRTHandler)
allHandlers.add('ND280Executable','TestSubmitter', RTHandler)
allHandlers.add('ND280Executable','Interactive', RTHandler)
allHandlers.add('ND280Executable','Batch', RTHandler)
allHandlers.add('ND280Executable','Cronus', RTHandler)
allHandlers.add('ND280Executable','Remote', LCGRTHandler)
allHandlers.add('ND280Executable','CREAM', LCGRTHandler)





__all__ = ['test_logger']

# this function may be used to test logging problems in runtime packages
def test_logger():
    import GangaCore.Utility.logging
    logger = GangaCore.Utility.logging.getLogger()
    # this is to test manually (visually) the logging level mechanism
    print "logger test RAW ",logger,logger.name
    logger.debug('logger test DEBUG')
    logger.info('logger test INFO')
    logger.warning('logger test WARNING')
    logger.error('logger test ERROR')
    logger.critical('logger test CRITICAL')


_this_junk_will_not_be_imported_to_GPI = None #because it is not listed in __all__ (and also it starts with underscore)




#
# Dictionary with all parameters available for the
# section GRID.
#
# 08/06/10 @ ubeda
#

from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Schema import *

class GRID(GangaObject):

    _comments = []
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#rb')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#proxy_server')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#role')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#group')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#dont_check_proxy')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#dont_check_myproxy')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#requirements')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#additional_jdl_parameters_')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#wms_service')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#max_cpu_time')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#max_wall_clock_time')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#ce_black_list')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#ce_white_list')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#se_black_list')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#se_white_list')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#remove_default_blacklist')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#virtual_organization')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#retry_count')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#shallow_retry_count')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#maxtarballsize')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#maxtarballsize')

    schemadic = {}
    schemadic['rb']                        = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[0])
    schemadic['proxy_server']              = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[1])
    schemadic['role']                      = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[2])
    schemadic['group']                     = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[3])
    schemadic['dont_check_proxy']          = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[4])
    schemadic['dont_check_myproxy']        = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[5])
    schemadic['requirements']              = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[6])
    schemadic['additional_jdl_parameters'] = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[7])
    schemadic['wms_service']               = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[8])
    schemadic['max_cpu_time']              = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[9])
    schemadic['max_wall_clock_time']       = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[10])
    schemadic['CE_black_list']             = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[11])
    schemadic['CE_white_list']             = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[12])
    schemadic['SE_black_list']             = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[13])
    schemadic['SE_white_list']             = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[14])
    schemadic['remove_default_blacklist']  = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[15])
    schemadic['virtual_organization']      = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[16])
    schemadic['retry_count']               = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[17])
    schemadic['shallow_retry_count']       = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[18])
    schemadic['maxtarballsize']            = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[19])
    schemadic['skipwmsauth']               = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[20])


    _schema =  Schema(Version(0,0), {})
    _hidden = 1



import os
import glob
import re
import shutil
from GangaCore.Core.exceptions import GangaFileError
from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Base.Proxy import getName
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.Utility.logging import getLogger
from fnmatch import fnmatch

logger = getLogger()
regex = re.compile('[*?\[\]]')


class IGangaFile(GangaObject):

    """IGangaFile represents base class for output files, such as MassStorageFile, LCGSEFile, DiracFile, LocalFile, etc 
    """
    _schema = Schema(Version(1, 1), {'namePattern': SimpleItem(
        defvalue="", doc='pattern of the file name')})
    _category = 'gangafiles'
    _name = 'IGangaFile'
    _hidden = 1
    __slots__ = list()

    def __init__(self):
        super(IGangaFile, self).__init__()

    def setLocation(self):
        """
        Sets the location of output files that were uploaded from the WN
        """
        raise NotImplementedError

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        raise NotImplementedError

    def get(self):
        """
        Retrieves locally all files that were uploaded before that 
        Order of priority about where a file is going to be placed are:
            1) The localDir as defined in the schema. (Exceptions thrown if this doesn't exist)
            2) The Job outpudir of the parent job if the localDir is not defined.
            3) raise an exception if neither are defined correctly.
        """
        if self.localDir:
            if not os.path.isdir(self.localDir):
                msg = "Folder '%s' doesn't exist. Please construct this before 'get'-ing a file." % self.localDir
                raise GangaFileError(msg)
            to_location = self.localDir
        else:
            try:
                to_location = self.getJobObject().outputdir
            except AssertionError:
                msg = "%s: Failed to get file object. Please set the `localDir` parameter and try again. e.g. file.localDir=os.getcwd();file.get()" % getName(self)
                logger.debug("localDir value: %s" % self.localDir)
                logger.debug("parent: %s" % self._getParent())
                raise GangaFileError(msg)

        # FIXME CANNOT perform a remote globbing here in a nice way so have to just perform a copy when dealing with wildcards
        if not os.path.isfile(os.path.join(to_location, self.namePattern)):
            returnable = self.copyTo(to_location)
            if not self.localDir:
                self.localDir = to_location
            return returnable
        else:
            logger.debug("File: %s already exists, not performing copy" % (os.path.join(to_location, self.namePattern), ))
            return True


    def getSubFiles(self, process_wildcards=False):
        """
        Returns the sub files if wildcards are used
        """
        # should we process wildcards? Used for inputfiles
        if process_wildcards:
            self.processWildcardMatches()

        # if we have subfiles, return that
        if hasattr(self, 'subfiles'):
            return self.subfiles

        return []

    def getFilenameList(self):
        """
        Returns the filenames of all associated files through a common interface
        """
        raise NotImplementedError

    def getWNScriptDownloadCommand(self, indent):
        """
        Gets the command used to download already uploaded file
        """
        raise NotImplementedError

    def put(self):
        """
        Postprocesses (upload) output file to the desired destination from the client
        """
        raise NotImplementedError

    def copyTo(self, targetPath):
        """
        Copy a the file to the local storage using the appropriate file-transfer mechanism
        This will raise an exception if targetPath isn't set to something sensible.
        Args:
            targetPath (str): Target path where the file is to copied to
        """
        if not isinstance(targetPath, str) and targetPath:
            raise GangaFileError("Cannot perform a copyTo with no given targetPath!")
        if regex.search(self.namePattern) is None\
            and os.path.isfile(os.path.join(self.localDir, self.namePattern)):

            if not os.path.isfile(os.path.join(targetPath, self.namePattern)):
                shutil.copy(os.path.join(self.localDir, self.namePattern), os.path.join(targetPath, self.namePattern))
            else:
                logger.debug("Already found file: %s" % os.path.join(targetPath, self.namePattern))
                
            return True

        # Again, cannot perform a remote glob here so have to ignore wildcards
        else:
            return self.internalCopyTo(targetPath)

    def internalCopyTo(self, targetPath):
        """
        Internal method for implementing the actual copy mechanism for each IGangaFile
        Args:
             targetPath (str): Target path where the file is to copied to
        """
        raise NotImplementedError

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        raise NotImplementedError

    def processWildcardMatches(self):
        """
        If namePattern contains a wildcard, populate the subfiles property
        """
        raise NotImplementedError

    def _auto_remove(self):
        """
        Remove called when job is removed as long as config option allows
        """
        self.remove()

    def _readonly(self):
        return False

    def _list_get__match__(self, to_match):
        if isinstance(to_match, str):
            return fnmatch(self.namePattern, to_match)
        # Note: type(DiracFile) = ObjectMetaclass
        # type(ObjectMetaclass) = type
        # hence checking against a class type not an instance
        if isinstance(type(to_match), type):
            return issubclass(self.__class__, to_match)
        return to_match == self

    def execSyscmdSubprocess(self, cmd):

        import subprocess

        exitcode = -999
        mystdout = ''
        mystderr = ''

        try:
            child = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            (mystdout, mystderr) = child.communicate()
            exitcode = child.returncode
        finally:
            pass

        return (exitcode, mystdout, mystderr)

    def remove(self):
        """
        Objects should implement something to overload this!
        """
        raise NotImplementedError

    def accessURL(self):
        """
        Return the URL including the protocol used to access a file on a certain storage element
        """
        raise NotImplementedError

    def hasMatchedFiles(self):
        """
        Return if this file has got valid matched files. Default implementation checks for
        subfiles and locations
        """

        # check for subfiles
        if (hasattr(self, 'subfiles') and len(self.subfiles) > 0):
            # we have subfiles so we must have actual files associated
            return True

        # check for locations
        if (hasattr(self, 'locations') and len(self.locations) > 0):
            return True

        return False

    def containsWildcards(self):
        """
        Return if the name has got wildcard characters
        """
        if regex.search(self.namePattern) is not None:
            return True

        return False

    def cleanUpClient(self):
        """
        This method cleans up the client space after performing a put of a file after a job has completed
        """

        # For all other file types (not LocalFile) The file in the outputdir is temporary waiting for Ganga to pass it to the storage solution
        job = self.getJobObject()

        for f in glob.glob(os.path.join(job.outputdir, self.namePattern)):
            try:
                os.remove(f)
            except OSError as err:
                if err.errno != errno.ENOENT:
                    logger.error('failed to remove temporary/intermediary file: %s' % f)
                    logger.debug("Err: %s" % err)
                    raise err

    def expandString(self, inputStr, fileName=''):
        """
        This method deals with the automatic string replacement in the string notation for IGangaFile objects
        Args:
            inputStr(str): This is the input string which is being evaluated/expanded
            fileName(str): This is an optional filename used to replace {fname}
        Returns:
            str:    This new string is the result of fully expanding the inputStr object
        """

        outputStr = inputStr

        if self._getParent() is not None:
            jobfqid = self.getJobObject().fqid
                                
            jobid = jobfqid
            subjobid = ''

            split = jobfqid.split('.')

            if len(split) > 1:
                jobid = split[0]
                subjobid = split[1]
          
            outputStr = outputStr.replace('{jid}', jobid)                                                                    
            outputStr = outputStr.replace('{sjid}', subjobid)

        if fileName:
            outputStr = outputStr.replace('{fname}', fileName)
        else:
            outputStr = outputStr.replace('{fname}', os.path.basename(self.namePattern))

        return outputStr




from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from GangaCore.Core.exceptions import GangaException
from GangaCore.GPIDev.Lib.Tasks.common import getJobByID
from GangaCore.GPIDev.Lib.Tasks.ITransform import ITransform
from GangaCore.GPIDev.Lib.Job.Job import JobError
from GangaCore.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.Core.exceptions import GangaAttributeError
from GangaLHCb.Lib.Tasks.LHCbUnit import LHCbUnit
from GangaCore.GPIDev.Base.Proxy import isType
from GangaLHCb.Lib.LHCbDataset.BKQuery import BKQuery
from GangaLHCb.Lib.LHCbDataset import LHCbDataset
from GangaDirac.Lib.Files.DiracFile import DiracFile
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Base.Proxy import stripProxy

logger = getLogger()

class LHCbTransform(ITransform):
    _schema = Schema(Version(1, 0), dict(ITransform._schema.datadict.items() + {
        'files_per_unit': SimpleItem(defvalue=-1, doc='Maximum number of files to assign to each unit from a given input dataset. If < 1, use all files.', typelist=["int"]),
        'splitter': ComponentItem('splitters', defvalue=None, optional=1, load_default=False, doc='Splitter to be used for units'),
        'queries': ComponentItem('query', defvalue=[], sequence=1, protected=1, optional=1, load_default=False, doc='Queries managed by this Transform'),
        'delete_chain_input': SimpleItem(defvalue=False, doc='Delete the Dirac input files/data after completion of each unit', typelist=["bool"]),
        'mc_num_units': SimpleItem(defvalue=0, doc="No. of units to create for MC generation"),

    }.items()))

    _category = 'transforms'
    _name = 'LHCbTransform'
    _exportmethods = ITransform._exportmethods + ['updateQuery', 'addQuery', 'removeUnusedData', 'cleanTransform']

    def __init__(self):
        super(LHCbTransform, self).__init__()

        # generally no delay neededd
        self.chain_delay = 0

    def addQuery(self, bk):
        """Add a BK query to this transform"""
        # Check if the BKQuery input is correct and append/update
        if not isType(bk, BKQuery):
            raise GangaAttributeError(
                None, 'LHCbTransform expects a BKQuery object passed to the addQuery method')

        # check we don't already have inputdata
        if len(self.queries) == 0 and len(self.inputdata) > 0:
            logger.error(
                "Cannot add both input data and BK queries. Input Data already present.")
            return

        # add the query and update the input data
        self.queries.append(bk)
        self.updateQuery()

    def addInputQuery(self, inDS):
        """Add the given input dataset to the list but only if BK queries aren't given"""
        if len(self.queries) > 0:
            logger.error(
                "Cannot add both input data and BK queries. Query already given")
            return

        super(LHCbTransform, self).addInputQuery(inDS)

    def cleanTransform(self):
        """Remove unused data and then unused jobs"""
        self.removeUnusedData()
        self.removeUnusedJobs()

    def removeUnusedData(self):
        """Remove any output data from orphaned jobs"""
        for unit in self.units:
            for jid in unit.prev_job_ids:
                try:
                    logger.warning("Removing data from job '%d'..." % jid)
                    job = getJobByID(jid)

                    jlist = []
                    if len(job.subjobs) > 0:
                        jlist = job.subjobs
                    else:
                        jlist = [job]

                    for sj in jlist:
                        for f in sj.outputfiles:
                            if isType(f, DiracFile) == "DiracFile" and f.lfn:
                                f.remove()
                except:
                    logger.error("Problem deleting data for job '%d'" % jid)
                    pass

    def createUnits(self):
        """Create new units if required given the inputdata"""

        # call parent for chaining
        super(LHCbTransform, self).createUnits()

        if len(self.inputdata) > 0:

            # check for conflicting input
            if self.mc_num_units > 0:
                logger.warning("Inputdata specified - MC Event info ignored")

            # loop over input data and see if we need to create any more units
            import copy
            for id, inds in enumerate(self.inputdata):

                if not isType(inds, LHCbDataset):
                    continue

                # go over the units and see what files have been assigned
                assigned_data = LHCbDataset()
                for unit in self.units:

                    if unit.input_datset_index != id:
                        continue

                    assigned_data.files += unit.inputdata.files

                # any new files
                new_data = LHCbDataset(files=self.inputdata[id].difference(assigned_data).files)

                if len(new_data.files) == 0:
                    continue

                # Create units for these files
                step = self.files_per_unit
                if step <= 0:
                    step = len(new_data.files)

                for num in range(0, len(new_data.files), step):
                    unit = LHCbUnit()
                    unit.name = "Unit %d" % len(self.units)
                    unit.input_datset_index = id
                    self.addUnitToTRF(unit)
                    unit.inputdata = copy.deepcopy(self.inputdata[id])
                    unit.inputdata.files = []
                    unit.inputdata.files += new_data.files[num:num+step]

        elif self.mc_num_units > 0:
            if len(self.units) == 0:
                # check for appropriate splitter
                from GangaLHCb.Lib.Splitters.GaussSplitter import GaussSplitter
                if not self.splitter or isType(self.splitter, GaussSplitter):
                    logger.warning("No GaussSplitter specified - first event info ignored")

                # create units for MC generation
                for i in range(0, self.mc_num_units):
                    unit = LHCbUnit()
                    unit.name = "Unit %d" % len(self.units)
                    self.addUnitToTRF(unit)
        else:
            import traceback
            traceback.print_stack()
            logger.error("Please specify either inputdata or MC info for unit generation")

    def createChainUnit(self, parent_units, use_copy_output=True):
        """Create an output unit given this output data"""

        # we need a parent job that has completed to get the output files
        incl_pat_list = []
        excl_pat_list = []
        for parent in parent_units:
            if len(parent.active_job_ids) == 0 or parent.status != "completed":
                return None

            for inds in self.inputdata:
                from GangaCore.GPIDev.Lib.Tasks.TaskChainInput import TaskChainInput
                if isType(inds, TaskChainInput) and inds.input_trf_id == parent._getParent().getID():
                    incl_pat_list += inds.include_file_mask
                    excl_pat_list += inds.exclude_file_mask

        # go over the output files and copy the appropriates over as input
        # files
        flist = []
        import re
        for parent in parent_units:
            job = getJobByID(parent.active_job_ids[0])
            if job.subjobs:
                job_list = job.subjobs
            else:
                job_list = [job]

            for sj in job_list:
                for f in sj.outputfiles:

                    # match any dirac files that are allowed in the file mask
                    if isType(f, DiracFile):
                        if len(incl_pat_list) > 0:
                            for pat in incl_pat_list:
                                if re.search(pat, f.lfn):
                                    flist.append("LFN:" + f.lfn)
                        else:
                            flist.append("LFN:" + f.lfn)

                        if len(excl_pat_list) > 0:
                            for pat in excl_pat_list:
                                if re.search(pat, f.lfn) and "LFN:" + f.lfn in flist:
                                    flist.remove("LFN:" + f.lfn)

        # just do one unit that uses all data
        unit = LHCbUnit()
        unit.name = "Unit %d" % len(self.units)
        unit.inputdata = LHCbDataset(files=[DiracFile(lfn=f) for f in flist])

        return unit

    def updateQuery(self, resubmit=False):
        """Update the dataset information of the transforms. This will
        include any new data in the processing or re-run jobs that have data which
        has been removed."""
        if len(self.queries) == 0:
            raise GangaException(
                None, 'Cannot call updateQuery() on an LHCbTransform without any queries')

        if self._getParent() is not None:
            logger.info('Retrieving latest bookkeeping information for transform %i:%i, please wait...' % (
                self._getParent().id, self.getID()))
        else:
            logger.info(
                'Retrieving latest bookkeeping information for transform, please wait...')

        # check we have an input DS per BK Query
        while len(self.queries) > len(self.inputdata):
            self.inputdata.append(LHCbDataset())

        # loop over the queries and add fill file lists
        for id, query in enumerate(self.queries):

            # Get the latest dataset
            latest_dataset = stripProxy(query.getDataset())

            # Compare to previous inputdata, get new and removed
            logger.info(
                'Checking for new and removed data for query %d, please wait...' % self.queries.index(query))
            dead_data = LHCbDataset()
            new_data = LHCbDataset()

            # loop over the old data and compare
            new_data.files += latest_dataset.difference(
                self.inputdata[id]).files
            dead_data.files += self.inputdata[
                id].difference(latest_dataset).files

            # for dead data, find then kill/remove any associated jobs
            # loop over units and check any associated with this DS
            # TODO: Follow through chained tasks
            for unit in self.units:
                # associted unit
                if unit.input_datset_index != id:
                    continue

                # find the job
                if len(unit.active_job_ids) == 0:
                    continue

                # check the data
                for f in dead_data.files:
                    if f in unit.inputdata.files:

                        # kill the job
                        job = getJobByID(unit.active_job_ids[0])
                        if job.status in ['submitted', 'running']:
                            job.kill()

                        # forget the job
                        unit.prev_job_ids.append(unit.active_job_ids[0])
                        unit.active_job_ids = []
                        break

            # in any case, now just set the DS files to the new set
            self.inputdata[id].files = []
            self.inputdata[id].files = latest_dataset.files



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: IApplication.py,v 1.1 2008-07-17 16:40:52 moscicki Exp $
##########################################################################

from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Schema import Schema, Version

import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()


class PostprocessStatusUpdate(Exception):

    """ This exception may be raised by postprocess hooks to change the job status."""

    __slots__ = ('status',)

    def __init__(self, status):
        Exception.__init__(self)
        self.status = status


class IApplication(GangaObject):

    """
     Base class for all application objects. Derived classes represent
     logical  applications  in  the  GPI and  implement  configuration
     handler functionality. The application configuration is the first
     phase of job submission.

     General rules for implementing the configure methods:

     In  general the  configure() and  master_configure()  methods are
     called always in  the context of job submission,  so in principle
     you may navigate to  the associated job object (including backend
     information). However it  is not advised to use  backend or extra
     sandbox  information at this  point.  Code  which depends  on the
     backend  should be  put in  application-specific  runtime handler
     which is the next step of job submission.

     The input/output dataset information may be used if neccessary.
     Objects in the job object tree should not be modified.

     Efficient implementation of splitting:

     If you want to enable the  typical case of splitting based on the
     dataset  (defined at  the  job  level) then  it  is very  simple:
     configure()  should only process  the inputdata  part of  the job
     configuration and master_configure() should  do the rest. In that
     case the splitter should not mutate the application object in the
     subjobs, because such changes will not be taken into account (and
     framework will have inconsistent behaviour).

     You  may  also take  an  extreme  approach  and move  the  entire
     application    configuration     to    configure().     Arbitrary
     modifications of the application  object in the subjobs which are
     done by  the splitter will  take effect.  But if  the application
     configuration  process is time  consuming it  will be  repeated a
     number of times which is inneficient.

     Otherwise  you   should  first   identify  which  parts   of  the
     application  object may be  altered by  the splitter  and process
     them  in  configure()   method.   The  master_configure()  should
     perform only  the time-consuming part of  the configuration which
     is shared among the subjobs.  This means that splitter should not
     try to  modify the application  parameters which are used  in the
     common master_configure().  """

    _schema = Schema(Version(0, 0), {})
    _category = 'applications'
    _hidden = 1

    __slots__ = list()

    def __init__(self):
        super(IApplication, self).__init__()

    def master_configure(self):
        """ Configure the shared/master  aspect of the application.
        Return a tuple (modified_flag,  appconfig).

        This method is always called exactly once, also in the case of
        splitting.

        Return tuple:

          - appconfig  (also  known   as  appextra)  is  an  arbitrary
            structure which will  be processed by application-specific
            runtime handler as the next step of job submission.

          - modified_flag  is True  if application  object  (self) has
            been modified during configure()

        If this method is not implemented in the derived class then it
        is ignored.  """

        return (0, None)

    def configure(self, master_appconfig):
        """ Configure  the specific aspect of the  application .  This
        method has a similar  meaning as master_configure() method and
        it should return a tuple (modified_flag,appconfig).

        This method must be  implemented in a derived class. Otherwise
        the submission will fail.

        Arguments:
          - master_appconfig is a result of the master job master_configure()
            method.

        In  case of  splitting this  method  will be  called for  each
        subjob  object  exactly  once  which  means that  it  will  be
        executed as many times as there are subjobs.

        If there  is no splitting  this method will be  called exactly
        once, after  master_configure().

        Transition from Ganga 4.0.x:
         - master_appconfig should be ignored (it is None anyway)
        """

        raise NotImplementedError

    def postprocess(self):
        """ Postprocessing after the job was reported as completed. By default do nothing.
        This method may raise an exception PostprocessStatusUpdate('failed'). In this
        case the job status will be 'failed'. The postprocess_failed() hook will NOT be called."""
        pass

    def postprocess_failed(self):
        """ Postprocessing after the job was reported as failed. By default do nothing."""
        pass

    def transition_update(self, new_status):
        """
        This method will be called just before the status of the parent Job changes to new_status.
        The default it to do nothing.
        """
        pass




##########################################################################
# Ganga - a computational task management tool for easy access to Grid resources
# http://cern.ch/ganga
#
# $Id: Coordinator.py,v 1.1.4.2 2009-07-14 14:44:17 ebke Exp $
#
# Copyright (C) 2003-2007 The Ganga Project
#
# This file is part of GangaCore.
#
# Ganga is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Ganga is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
##########################################################################

"""
 Internal services coordinator :
  takes care of conditional enabling/disabling of internal services (job monitoring loop, job registry, job
  repository/workspace) when credentials become invalid preventing normal functioning of these services.
  E.g: invalid grid proxy triggers the monitor-loop stop
"""
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Base.Proxy import getName

from GangaCore.GPIDev.Credentials import credential_store
from GangaCore.GPIDev.Credentials.AfsToken import AfsToken

log = getLogger()

# the overall state of Ganga internal services
servicesEnabled = True


def _diskSpaceChecker():
    """
    the callback function used internally by Monitoring Component
    Reads and calls the checking function provided in the configuration. 
    If this checking function returns "False" the internal services are disabled making Ganga read-only:
    e.g:
    [PollThread]
    DiskSpaceChecker =  
        import commands
        diskusage = commands.getoutput('df -l -P %s/workspace' % config['Configuration']['gangadir'])
        used  = diskusage.splitlines()[1].split()[4] # get disk usage (in %)
        return int(used[:-1])<70
    """
    log.debug("Checking disk space")
    try:
        config = getConfig('PollThread')

        if config['DiskSpaceChecker']:
            _checker = lambda: True
            try:
                # create the checker
                from GangaCore.Runtime import _prog
                import new
                ns = {}
                code = "def check():"
                for line in config['DiskSpaceChecker'].splitlines():
                    code += "\t%s\n" % line
                exec code in ns
                _checker = new.function(
                    ns["check"].__code__, _prog.local_ns, 'check')
            except Exception as e:
                log.warning(
                    'Syntax errors in disk space checking code: %s. See [PollThread]DiskSpaceChecker' % e)
                return False

            # call the checker
            if _checker() is False:
                disableInternalServices()
                log.warning('You are running out of disk space! '
                            'To protect against possible write errors all internal services has been disabled.'
                            'If you believe the problem has been solved type "reactivate()" to re-enable '
                            'interactions within this session.')
    except Exception as msg:
        log.warning(
            'Exception in free disk space checking code: %s. See [PollThread]DiskSpaceChecker' % msg)
        return False
    return True


def disableMonitoringService():

    # disable the mon loop
    log.debug("Shutting down the main monitoring loop")
    from GangaCore.Core.MonitoringComponent.Local_GangaMC_Service import _purge_actions_queue, stop_and_free_thread_pool
    _purge_actions_queue()
    stop_and_free_thread_pool()
    log.debug("Disabling the central Monitoring")
    from GangaCore.Core import monitoring_component
    monitoring_component.disableMonitoring()


def disableInternalServices():
    """
    Deactivates all the internal services :
          * monitoring loop
          * registry/repository and workspace (or GPI entierly)
    Currently this method is called whenever:
          * one of the managed credentials (AFS token or Grid Proxy) is detected as beeing *invalid* by the monitoring component
          * the user is running out of space
    """

    log.info("Ganga is now attempting to shut down all running processes accessing the repository in a clean manner")
    log.info(" ... Please be patient! ")


    ## MOVED TO THE END OF THE SHUTDOWN SO THAT WE NEVER ACCESS A REPO BEFORE WE ARE FINISHED!
    # flush the registries
    #log.debug("Coordinator Shutting Down Repository_runtime")
    #from GangaCore.Runtime import Repository_runtime
    #Repository_runtime.shutdown()

    global servicesEnabled

    if not servicesEnabled:
        log.error("Cannot disable services, they've already been disabled")
        from GangaCore.Core.exceptions import GangaException
        raise GangaException("Cannot disable services")

    log.debug("Disabling the internal services")

    # disable the mon loop
    disableMonitoringService()

    # For debugging what services are still alive after being requested to stop before we close the repository
    #from GangaCore.Core.MonitoringComponent.Local_GangaMC_Service import getStackTrace
    # getStackTrace()
    # log.info(queues_threadpoolMonitor._display(0))

    log.debug("Ganga is now about to shutdown the repository, any errors after this are likely due to badly behaved services")

    log.info("Ganga is shutting down the repository, to regain access, type 'reactivate()' at your prompt")

    # flush the registries
    #log.debug( "Coordinator Shutting Down Repository_runtime" )
    #from GangaCore.Runtime import Repository_runtime
    # Repository_runtime.shutdown()

    # this will disable any interactions with the registries (implicitly with
    # the GPI)
    servicesEnabled = False


def enableMonitoringService():
    from GangaCore.Core import monitoring_component
    monitoring_component.alive = True
    monitoring_component.enableMonitoring()
    from GangaCore.Core.MonitoringComponent.Local_GangaMC_Service import _makeThreadPool, ThreadPool
    if not ThreadPool or len(ThreadPool) == 0:
        _makeThreadPool()
    global servicesEnabled
    servicesEnabled = True

def enableInternalServices():
    """
    activates the internal services previously disabled due to expired credentials
    """
    global servicesEnabled

    if servicesEnabled:
        log.error("Cannot (re)enable services, they're already running")
        from GangaCore.Core.exceptions import GangaException
        raise GangaException("Cannot (re)enable services")

    # startup the registries
    from GangaCore.Runtime import Repository_runtime
    Repository_runtime.bootstrap()

    # make sure all required credentials are valid
    invalid_afs = [afsToken for afsToken in credential_store.get_all_matching_type(AfsToken()) if not afsToken.is_valid()]

    if invalid_afs:
        log.error('No valid AFS token was found. Please re-authorise before reactivating this session.')
        return

    log.debug("Enabling the internal services")
    # re-enable the monitoring loop as it's been explicityly requested here
    enableMonitoringService()

    servicesEnabled = True
    log.info('Internal services reactivated successfuly')


def checkInternalServices(errMsg='Internal services disabled. Job registry is read-only.'):
    """
    Check the state of internal services and return a ReadOnlyObjectError exception
    in case the state is disabled.    
    """

    global servicesEnabled
    from GangaCore.GPIDev.Base import ReadOnlyObjectError

    if not servicesEnabled:
        raise ReadOnlyObjectError(errMsg)

#
#$Log: not supported by cvs2svn $
# Revision 1.1.4.1  2009/07/08 11:18:21  ebke
# Initial commit of all - mostly small - modifications due to the new GangaRepository.
# No interface visible to the user is changed
#
# Revision 1.1  2008/07/17 16:40:50  moscicki
# migration of 5.0.2 to HEAD
#
# the doc and release/tools have been taken from HEAD
#
# Revision 1.3.6.4  2008/03/11 15:22:42  moscicki
# merge from Ganga-5-0-restructure-config-branch
#
# Revision 1.3.6.3.2.1  2008/03/07 13:36:07  moscicki
# removal of [DefaultJobRepository] and [FileWorkspace]
# new options in [Configuration] user, gangadir, repositorytype, workspacetype
#
# Revision 1.3.6.3  2008/02/12 09:25:52  amuraru
# fixed repositories shutdown
#
# Revision 1.3.6.2  2008/02/05 12:33:23  amuraru
# fixed DiskSpaceChecker alignment
#
# Revision 1.3.6.1  2007/12/10 19:24:42  amuraru
# merged changes from Ganga 4.4.4
#
# Revision 1.7  2007/12/05 12:42:54  amuraru
# Ganga/Core/InternalServices/Coordinator.py
#
# Revision 1.6  2007/12/04 12:59:03  amuraru
#*** empty log message ***
#
# Revision 1.5  2007/11/26 14:04:30  amuraru
# allow indentation using \t tab char in DiskSpaceChecker
#
# Revision 1.4  2007/10/29 14:06:00  amuraru
# added free disk space checker to monitoring loop
#
# Revision 1.3  2007/07/27 18:02:34  amuraru
# updated to comply with the latest requirement for GPI free functions docstrings
#
# Revision 1.2  2007/07/27 14:31:56  moscicki
# credential and clean shutdown updates from Adrian (from Ganga-4-4-0-dev-branch)
#
# Revision 1.1.2.1  2007/07/27 13:04:00  amuraru
#*** empty log message ***
#
#




import random
import tempfile
import string

def generate_unique_temp_file(ext = '.txt'):
    """ Generate a unique file with a given filename with some random contents and return the name of the file on disk
    Args:
        ext (str): This is the extension (including '.') to give to the file of interest
    """

    with tempfile.NamedTemporaryFile(mode='w', suffix=ext, delete=False) as this_file:

        this_file.write(''.join(random.choice(string.ascii_uppercase+string.digits) for _ in range(20)))

        return this_file.name





##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Merger.py,v 1.5 2009-03-18 10:46:01 wreece Exp $
##########################################################################

from GangaCore.GPIDev.Adapters.IPostProcessor import PostProcessException
from GangaCore.GPIDev.Adapters.IMerger import IMerger
from GangaCore.GPIDev.Schema import FileItem, SimpleItem
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.GPIDev.Lib.File.File import File
from GangaCore.GPIDev.Lib.File.LocalFile import LocalFile
from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile
from GangaCore.Utility.Config import ConfigError, getConfig
from GangaCore.Utility.Plugin import allPlugins
from GangaCore.Utility.logging import getLogger
import tarfile
import commands
import tempfile
import os
import shutil
import string
import copy

logger = getLogger()

class GaudiExecMerger(IMerger):

    """Merger class for GaudiExec jobs

    GaudiExec merger will use the version of hadd in your GaudiExec application
    to merge your files for maximum compatibility.

    Usage:

    gm = GaudiExecMerger()
    gm.files = ['hist.root','trees.root']
    gm.overwrite = True #False by default
    gm.ignorefailed = True #False by default
    gm.args = '-f2' #pass arguments to hadd

    # will produce the specified files
    j = Job() 
    j.outputfiles = ['hist.root','trees.root']
    j.splitter = SomeSplitter()
    j.postprocessors = gm
    j.submit()

    The merge object will be used to merge the output of
    each subjob into j.outputdir. This will be run when
    the job completes. If the ignorefailed flag has been set
    then the merge will also be run as the job enters the
    killed or failed states.

    The above merger object can also be used independently
    to merge a list of jobs or the subjobs of an single job.

    #gm defined above
    gm.merge(j, outputdir = '~/merge_dir')
    gm.merge([.. list of jobs ...], '~/merge_dir', ignorefailed = True, overwrite = False)

    If ignorefailed or overwrite are set then they override the
    values set on the merge object.

    A summary of all the files merged will be created for each entry in files.
    This will be created when the merge of those files completes
    successfully. The name of this is the same as the output file, with the
    '.merge_summary' extension appended and will be placed in the same directory
    as the merge results.

    If outputdir is not specified, the default location specfied
    in the [Mergers] section of the .gangarc file will be used.

    The way this works is to copy the compressed cmake input sandbox from the sharedir to your tmp area,
    untar it and execute ./run hadd in order to make use of the correct environment.

    """

    _category = 'postprocessor'
    _name = 'GaudiExecMerger'
    _schema = IMerger._schema.inherit_copy()
    _schema.datadict['args'] = SimpleItem(defvalue=None, doc='Arguments to be passed to hadd.',
                                          typelist=[str, None])

    def merge(self, jobs, outputdir=None, ignorefailed=None, overwrite=None):

        if ignorefailed is None:
            ignorefailed = self.ignorefailed

        if overwrite is None:
            overwrite = self.overwrite

        from GangaCore.GPIDev.Lib.Job import Job

        if not outputdir:
            outputdir = getDefaultMergeDir()
        else:
            if isType(outputdir, Job):
                # use info from job
                outputdir = outputdir.outputdir
            else:
                outputdir = os.path.expanduser(outputdir)

        files = {}

        if isType(jobs, Job):
            if outputdir is None:
                outputdir = jobs.outputdir
            return self.merge(jobs.subjobs, outputdir=outputdir, ignorefailed=ignorefailed, overwrite=overwrite)

        if not len(jobs):
            logger.warning('The jobslice given was empty. The merge will not continue.')
            return self.success
        masterjob = jobs[0].master

        for j in jobs:
            # first check that the job is ok
            if j.status != 'completed':
                # check if we can keep going
                if j.status == 'failed' or j.status == 'killed':
                    if ignorefailed:
                        logger.warning('Job %s has status %s and is being ignored.', j.fqid, j.status)
                        continue
                    else:
                        raise PostProcessException('Job %s has status %s and so the merge can not continue. '
                                                   'This can be overridden with the ignorefailed flag.' % (j.fqid, j.status))
                else:
                    raise PostProcessException("Job %s is in an unsupported status %s and so the merge can not continue. '\
                    'Supported statuses are 'completed', 'failed' or 'killed' (if the ignorefailed flag is set)." % (j.fqid, j.status))

            if len(j.subjobs):
                sub_result = self.merge(
                    j.subjobs, outputdir=j.outputdir, ignorefailed=ignorefailed, overwrite=overwrite)
                if (sub_result == self.failure) and not ignorefailed:
                    raise PostProcessException('The merge of Job %s failed and so the merge can not continue. '
                                               'This can be overridden with the ignorefailed flag.' % j.fqid)

            import glob
            for f in self.files:

                for matchedFile in glob.glob(os.path.join(j.outputdir, f)):
                    relMatchedFile = ''
                    try:
                        relMatchedFile = os.path.relpath(
                            matchedFile, j.outputdir)
                    except Exception as err:
                        logger.debug("Err: %s" % err)
                        GangaCore.Utility.logging.log_unknown_exception()
                        relMatchedFile = relpath(matchedFile, j.outputdir)
                    if relMatchedFile in files:
                        files[relMatchedFile].append(matchedFile)
                    else:
                        files[relMatchedFile] = [matchedFile]

                if not len(glob.glob(os.path.join(j.outputdir, f))):
                    if ignorefailed:
                        logger.warning('The file pattern %s in Job %s was not found. The file will be ignored.', f, j.fqid)
                        continue
                    else:
                        raise PostProcessException('The file pattern %s in Job %s was not found and so the merge can not continue. '
                                                   'This can be overridden with the ignorefailed flag.' % (f, j.fqid))
                # files[f].extend(matchedFiles)

        for k in files.keys():
            # make sure we are not going to over write anything
            outputfile = os.path.join(outputdir, k)
            if os.path.exists(outputfile) and not overwrite:
                raise PostProcessException('The merge process can not continue as it will result in over writing. '
                                           'Either move the file %s or set the overwrite flag to True.' % outputfile)

            # make the directory if it does not exist
            if not os.path.exists(outputdir):
                os.makedirs(outputdir)

            # recreate structure from output sandbox
            outputfile_dirname = os.path.dirname(outputfile)
            if outputfile_dirname != outputdir:
                if not os.path.exists(outputfile_dirname):
                    os.mkdir(outputfile_dirname)

            # check that we are merging some files
            if not files[k]:
                logger.warning('Attempting to merge with no files. Request will be ignored.')
                continue

            # check outputfile != inputfile
            for f in files[k]:
                if f == outputfile:
                    raise PostProcessException(
                        'Output file %s equals input file %s. The merge will fail.' % (outputfile, f))
            # merge the lists of files with a merge tool into outputfile
            msg = None
            try:
                self.mergefiles(masterjob, files[k], outputfile)

                # create a log file of the merge
                # we only get to here if the merge_tool ran ok
                log_file = '%s.merge_summary' % outputfile
                with open(log_file, 'w') as log:
                    log.write('# -- List of files merged -- #\n')
                    for f in files[k]:
                        log.write('%s\n' % f)
                    log.write('# -- End of list -- #\n')

            except PostProcessException as e:
                msg = str(e)

                # store the error msg
                log_file = '%s.merge_summary' % outputfile
                with open(log_file, 'w') as log:
                    log.write('# -- Error in Merge -- #\n')
                    log.write('\t%s\n' % msg)
                raise e

        return self.success


    def mergefiles(self, masterjob, file_list, output_file):

        #First grab the job object etc.
        j = masterjob
        sharedir = j.application.is_prepared.path()
        tmp_dir = tempfile.gettempdir()
        #Extract the run script
        tar = tarfile.open(os.path.join(sharedir, "cmake-input-sandbox.tgz"), "r:gz")
        tar.extractall(tmp_dir)
        #Run the hadd command from the application environment.
        merge_cmd = os.path.join(tmp_dir, 'run')
        default_arguments = '-f'
        merge_cmd += ' hadd '
        if self.args:  # pass any args on
            merge_cmd += ' %s ' % self.args

        # don't add a -f unless needed
        if not default_arguments in merge_cmd:
            merge_cmd += ' %s ' % default_arguments

        # add the list of files, output file first
        arg_list = [output_file]
        arg_list.extend(file_list)
        merge_cmd += string.join(arg_list, ' ')

        rc, out = commands.getstatusoutput(merge_cmd)

        try:
            #Clean up - first make a list of everything in the tarfile
            tarlist = tar.getnames()
            #Pop the run script and find the common prefix of the application folder
            tarlist.pop(tarlist.index('run'))
            folderToRemove = os.path.commonprefix(tarlist)
            #Now remove the files
            os.remove(os.path.join(tmp_dir, 'run'))
            shutil.rmtree(os.path.join(tmp_dir, folderToRemove))
        except OSError:
            logger.error('Failed to remove temporary files from merging at %s' % tmp_dir)


        log_file = '%s.hadd_output' % output_file
        with open(log_file, 'w') as log:
            log.write('# -- Hadd output -- #\n')
            log.write('%s\n' % out)

        if rc:
            logger.error(out)
            raise PostProcessException(
                'The ROOT merge failed to complete. The command used was %s.' % merge_cmd)
